# -*- coding: utf-8 -*-

"""Provide access to IoT-LAB experiments as MQTT agents"""

from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

__version__ = '0.6.1'
